// This file is part of VKF, a CPython library
// for Machine Learning based on Lattice Theory.
//
// Copyright (c) 2020 Dmitry V. Vinogradov
//
// This Source Code Form is subject to the terms of the X11 (MIT)
// License. If a copy of the X11 License was not distributed
// with this file, You can obtain one at https://opensource.org/licenses/MIT/.

#include "Induction.hpp"

void* vkf::Induction::GenerateNext() {
    
    while (!m_StopRequested) {
        m_HypothesesNumber++;
        if (m_HypothesesNumber >= m_hypotheses_to_add)
            m_StopRequested = true;
        HypothesisPtr pointer_to_hypothesis(new vkf::Hypothesis(m_fContext,m_obstacles));
        std::lock_guard<std::mutex> modification_lock(m_ModificationMutex);
        this->emplace_back(std::move(pointer_to_hypothesis));
    }
    return(nullptr);
}

unsigned int vkf::Induction::DetectSuspicious() {
	unsigned int result = 0;
	boost::dynamic_bitset<>::size_type position1, position2; // positions of first and second parents of a suspicious hypothesis 
	bool first_parent_has_a_real_cause = false;
	bool both_parents_have_different_causes = false;
	for (HypothesesList::iterator it = this->begin(); it != this->end(); ++it) {
		/* look for hypotheses with only 2 parents */
		if ((*it)->GetParentsNumber() == 2) {
			/* determine positions of both parents of given hypothesis */ 		
			position1 = (*it)->FindFirst();
			position2 = (*it)->FindNext(position1);				
			/* does the first parent have a "real" cause? */
			first_parent_has_a_real_cause = false;
			for (HypothesesList::const_iterator cit = this->cbegin(); cit != this->cend(); cit++) {
				if ((*cit)->GetParentsNumber() > 2 && ((*cit)->IsParent(position1) && !(*cit)->IsParent(position2))) {
					first_parent_has_a_real_cause = true;
					break;
				}
			}
			if (first_parent_has_a_real_cause) {
		    	/* does the second parent have a "real" cause? */
				both_parents_have_different_causes = false;
				for (HypothesesList::const_iterator cit = this->cbegin(); cit != this->cend(); cit++){
					if ((*cit)->GetParentsNumber() > 2 && (!(*cit)->IsParent(position1) && (*cit)->IsParent(position2))) {
						both_parents_have_different_causes = true;
						break;
					}
				}
				if (both_parents_have_different_causes) {
					(*it)->SetSuspiciousStatus(true);
					++result;
				}
			}
		}
	}
	return result;
}

void vkf::Induction::RestoreStatus() {
	for (HypothesesList::iterator it = this->begin(); it != this->end(); ++it) {
		(*it)->SetSuspiciousStatus(false);
	}
	return;
}

namespace vkf {
    #ifndef RANDOM_GENERATOR
    #define RANDOM_GENERATOR
    std::mt19937_64 AOMatrix::mt_random_engine;
    std::uniform_int_distribution<unsigned long long> AOMatrix::index_distribution;
    #endif // RANDOM_GENERATOR

    //    Exporting `vkf::Induction` variable with alias name `causes`
    //    (Has the same effect as `BOOST_DLL_ALIAS(vkf::Induction::causes, causes)`)
    API Induction causes;
    Induction causes;
} // namespace vkf
