// This file is part of VKF, a CPython library
// for Machine Learning based on Lattice Theory.
//
// Copyright (c) 2020 Dmitry V. Vinogradov
//
// This Source Code Form is subject to the terms of the X11 (MIT)
// License. If a copy of the X11 License was not distributed
// with this file, You can obtain one at https://opensource.org/licenses/MIT/.

#ifndef VKF_QUALIFY_HPP
#define VKF_QUALIFY_HPP

#include <mariadb++/account.hpp>
#include <mariadb++/connection.hpp>

#include "Verges.hpp"

namespace vkf {

class Qualify : public std::vector<double> {
    private:
        std::string attribute_name;
        double maximum;
        double minimum;
    public:
        std::string GetAttrName() const { return attribute_name; };
        ~Qualify() = default;
        Qualify(const std::string name) : attribute_name(name), maximum(0.0), minimum(0.0) {};
        void SetMaximum(double num) { maximum = num; return; };
        void SetMinimum(double num) { minimum = num; return; };
        double GetMaximum() const { return maximum; };
        double GetMinimum() const { return minimum; };
        Qualify(const Verges &temp, const std::string name)
            : attribute_name(name), maximum(temp.GetMaximum()), minimum(temp.GetMinimum()) {
            for (auto value : temp) {
                this->push_back(value);
            }
        };

        boost::dynamic_bitset<> Encode(double value) const {
            std::uint32_t total = this->size();
            boost::dynamic_bitset<> bdbs;
            for (std::uint32_t ndx = 0; ndx < total; ndx++) {
                if (value >= at(ndx)) {
                    bdbs.push_back(true);
                } else {
                    bdbs.push_back(false);
                }
            }
            for (std::uint32_t ndx = 0; ndx < total; ndx++) {
                if (value < at(ndx)) {
                    bdbs.push_back(true);
                } else {
                    bdbs.push_back(false);
                }
            }
            return bdbs;
        };

        boost::dynamic_bitset<> Encode(double low_bound, double high_bound) const {
            std::uint32_t total = this->size();
            boost::dynamic_bitset<> bdbs;
            for (std::uint32_t ndx = 0; ndx < total; ndx++) {
                if (low_bound >= at(ndx)) {
                    bdbs.push_back(true);
                } else {
                    bdbs.push_back(false);
                }
            }
            for (std::uint32_t ndx = 0; ndx < total; ndx++) {
                if (high_bound < at(ndx)) {
                    bdbs.push_back(true);
                } else {
                    bdbs.push_back(false);
                }
            }
            return bdbs;
        };


        std::pair<double,double> Decode(const boost::dynamic_bitset<>& bdbs) const {
            std::uint32_t total = this->size();
            assert (bdbs.size() == 2*total);
            double low_bound = this->minimum;
            double high_bound = this->maximum;
            for (std::uint32_t ndx = 0; ndx < total; ndx++) {
                if (bdbs[ndx] == true) {
                    low_bound = at(ndx);
                }
            }
            for (std::uint32_t ndx = 2*total-1; ndx >= total; ndx--) {
                if (bdbs[ndx] == true) {
                    high_bound = at(ndx-total);
                }
            }
            return std::pair<double,double>(low_bound, high_bound);
        };

};

class Qualifier : public std::vector<Qualify> {
    public:
        ~Qualifier() = default;
        Qualifier(  const std::string &verges_table_name,
                    const std::string &database,
                    const std::string &db_host = DEFAULT_HOST,
                    const std::string &db_user = DEFAULT_USER,
                    const std::string &db_pswd = DEFAULT_PASS
                ) {
            // set up the account
            mariadb::account_ref acc = mariadb::account::create(db_host, db_user, db_pswd);
            // create connection
            mariadb::connection_ref con = mariadb::connection::create(acc);
            con->execute("USE " + database);
            // obtain content of table encoder
            mariadb::result_set_ref result = con->query("SELECT * FROM " + verges_table_name);
            con->~connection();
            acc->~account();
            // reading from a result set
            while (result->next()) {
                std::string attr_name = result->get_string("attribute_name"); /**< This variable contains a name of the attribute */
                auto it = std::find_if(this->begin(), this->end(),
                    [&](const Qualify &arg){ return arg.GetAttrName() == attr_name; });
                if (it == this->end()) {
                    this->emplace_back(attr_name);
                    this->back().push_back(result->get_double("value_verge"));
                } else {
                    it->push_back(result->get_double("value_verge"));
                }
            }
            for (auto it = this->begin(); it != this->end(); it++) {
                if (it->size() > 2) {
                    std::list<double> temp;
                    for (std::uint32_t ndx = 0; ndx < it->size(); ndx++) {
                        temp.push_back(it->at(ndx));
                    }
                    temp.sort();
                    it->SetMaximum(temp.back());
                    temp.pop_back();
                    it->SetMinimum(temp.front());
                    temp.pop_front();
                    it->clear();
                    for (double num : temp) {
                        it->push_back(num);
                    }
                } else {
                    it->SetMaximum(it->back());
                    it->SetMinimum(it->front());
                }
            }
        };
};

} // namespace vkf

#endif // VKF_QUALIFY_HPP
