// This file is part of VKF, a CPython library
// for Machine Learning based on Lattice Theory.
//
// Copyright (c) 2020 Dmitry V. Vinogradov
//
// This Source Code Form is subject to the terms of the X11 (MIT)
// License. If a copy of the X11 License was not distributed
// with this file, You can obtain one at https://opensource.org/licenses/MIT/.

#ifndef VKF_SAMPLE_HPP
#define VKF_SAMPLE_HPP

#define DEFAULT_HOST "localhost"
#define DEFAULT_PASS "toor"
#define DEFAULT_USER "root"

#include <cassert>
#include <utility> // for std::pair
#include <string>
#include <vector>
#include "Encoder.hpp"
//#include "Join.hpp"
#include "Qualify.hpp"
#include "Beget.hpp"

namespace vkf{
class Example : public boost::dynamic_bitset<boost::dynamic_bitset<>::block_type> {
//class BOOST_SYMBOL_EXPORT Encoded : public boost::dynamic_bitset<boost::dynamic_bitset<>::block_type> {
    protected:
    private:
        bool m_sign; //!< Member variable "m_sign" indicates the goal property occurence
    public:
        /** Default constructor  c++11 type definition */
        Example() = default;
        /** copy-constructor */
        Example(const Example&) = delete;
        /** move-constructor */
        Example(Example&&) = default;
        /** copy-operator= */
        Example& operator=(const Example&) = delete;
        /** move-operator= */
        Example& operator=(Example&&) = default;
        /** Constructor Example from bool and  boost::dynamic_bitset<> implementation */
        Example(bool real_sign, boost::dynamic_bitset<>&& encoded_bitset) :
                   boost::dynamic_bitset<boost::dynamic_bitset<>::block_type>(encoded_bitset), m_sign(real_sign) {};
        /** default destructor implementation */
        ~Example() = default;
        /** Access m_sign
        * \return the bool value of m_sign */
        bool GetSign() const { return m_sign; }
};

typedef std::unique_ptr<Example> ExamplePtr;

//    class BOOST_SYMBOL_EXPORT Sample : public /* virtual */ ExampleList {
class Sample : public /* virtual */ std::vector<ExamplePtr> {
    protected:
    private:
    public:
        Sample() = default;
        Sample(Sample&& rhs) = default;
        Sample(const Sample& rhs) = default;
        /** copy-operator= */
        Sample& operator=(const Sample& rhs) = delete;
        /** move-operator= */
        Sample& operator=(Sample&& rhs) = default;
        /** Constructor Sample for discrete attributes only */
        Sample(const Encoder& encoder,
                const std::string &data_table_name,
                const std::string &database,
                const std::string &db_host = DEFAULT_HOST,
                const std::string &db_user = DEFAULT_USER,
                const std::string &db_pswd = DEFAULT_PASS
                ) {
            // set up the account
            mariadb::account_ref acc = mariadb::account::create(db_host, db_user, db_pswd);
            // create connection
            mariadb::connection_ref con = mariadb::connection::create(acc);
            // open database
            con->execute("USE " + database);
            // obtain content of table with data
            mariadb::result_set_ref result = con->query("SELECT * FROM " + data_table_name);
            con->~connection();
            acc->~account();
            // reading from a result set
            std::size_t sign_pos = result->column_index("real_sign");
            while (result->next()) {
                bool real_sign = result->get_boolean(sign_pos); /**< This variable contains the real sign of the object */
                boost::dynamic_bitset<> encoded_bitset;
                for (auto maps : encoder) {
                    std::string attribute_name = maps.getAttributeName(); /**< This variable contains the name of an attribute in the current column */
                    std::size_t pos = result->column_index(attribute_name);
                    std::string attribute_value = result->get_string(pos); /**< This variable saves the value of the attribute */
                    boost::dynamic_bitset<> bdbs = maps.findEncoding(attribute_value);
                    for (std::size_t ndx = 0; ndx < bdbs.size(); ndx++ ) {
                        encoded_bitset.push_back(bdbs[ndx]);
                    }
                }
                this->emplace_back(std::make_unique<Example>(real_sign, std::move(encoded_bitset)));
            }
        }
        /** Constructor Sample for continuous attributes only */
        Sample( const Qualifier &thresholds,
                const Beget &generator,
                const std::string &data_table_name,
                const std::string &database,
                const std::string &db_host = DEFAULT_HOST,
                const std::string &db_user = DEFAULT_USER,
                const std::string &db_pswd = DEFAULT_PASS
                ) {
            // set up the account
            mariadb::account_ref acc = mariadb::account::create(db_host, db_user, db_pswd);
            // create connection
            mariadb::connection_ref con = mariadb::connection::create(acc);
            // open database
            con->execute("USE " + database);
            // obtain content of table with data
            mariadb::result_set_ref result = con->query("SELECT * FROM " + data_table_name);
            con->~connection();
            acc->~account();
            // reading from a result set
            std::size_t sign_pos = result->column_index("real_sign");
            while (result->next()) {
                bool real_sign = result->get_boolean(sign_pos); /**< This variable contains the real sign of the object */
                boost::dynamic_bitset<> encoded_bitset;
                for (auto verges : thresholds) {
                    std::string attr_name = verges.GetAttrName(); /**< This variable contains the name of an attribute in the current column */
                    std::size_t pos = attr_name.find("__");
                    if (pos == std::string::npos) /* an original attribute */ {
                        boost::dynamic_bitset<> bdbs = verges.Encode(result->get_double(attr_name));
                        for (std::size_t indx = 0; indx < bdbs.size(); indx++ ) {
                            encoded_bitset.push_back(bdbs[indx]);
                        }
                    } else /* a pair of original attributes */ {
                        std::string attr1_name = attr_name.substr(0, pos);
                        std::string attr2_name = attr_name.substr(pos + 2); // 2 == std::string("__").size()
                        boost::dynamic_bitset<> bdbs = verges.Encode(
                            generator.Estimate(result->get_double(attr1_name), result->get_double(attr2_name), attr1_name, attr2_name));
                        for (std::size_t indx = 0; indx < bdbs.size(); indx++ ) {
                            encoded_bitset.push_back(bdbs[indx]);
                        }
                    }
                }
                this->emplace_back(std::make_unique<Example>(real_sign, std::move(encoded_bitset)));
            }
        }

        /** Constructor Sample for both types attributes */
        Sample( const Encoder& encoder,
                const Qualifier &thresholds,
                const Beget &generator,
                const std::string &data_table_name,
                const std::string &database,
                const std::string &db_host = DEFAULT_HOST,
                const std::string &db_user = DEFAULT_USER,
                const std::string &db_pswd = DEFAULT_PASS
                ) {
            // set up the account
            mariadb::account_ref acc = mariadb::account::create(db_host, db_user, db_pswd);
            // create connection
            mariadb::connection_ref con = mariadb::connection::create(acc);
            // open database
            con->execute("USE " + database);
            // obtain content of table with data
            mariadb::result_set_ref result = con->query("SELECT * FROM " + data_table_name);
            con->~connection();
            acc->~account();
            // reading from a result set
            std::size_t sign_pos = result->column_index("real_sign");
            while (result->next()) {
                bool real_sign = result->get_boolean(sign_pos); /**< This variable contains the real sign of the object */
                boost::dynamic_bitset<> encoded_bitset;
                for (auto maps : encoder) {
                    std::string attribute_name = maps.getAttributeName(); /**< This variable contains the name of an attribute in the current column */
                    std::size_t pos = result->column_index(attribute_name);
                    std::string attribute_value = result->get_string(pos); /**< This variable saves the value of the attribute */
                    boost::dynamic_bitset<> bdbs = maps.findEncoding(attribute_value);
                    for (std::size_t ndx = 0; ndx < bdbs.size(); ndx++ ) {
                        encoded_bitset.push_back(bdbs[ndx]);
                    }
                }
                for (auto verges : thresholds) {
                    std::string attr_name = verges.GetAttrName(); /**< This variable contains the name of an attribute in the current column */
                    std::size_t pos = attr_name.find("__");
                    if (pos == std::string::npos) /* an original attribute */ {
                        boost::dynamic_bitset<> bdbs = verges.Encode(result->get_double(attr_name));
                        for (std::size_t indx = 0; indx < bdbs.size(); indx++ ) {
                            encoded_bitset.push_back(bdbs[indx]);
                        }
                    } else /* a pair of original attributes */ {
                        std::string attr1_name = attr_name.substr(0, pos);
                        std::string attr2_name = attr_name.substr(pos + 2); // 2 == std::string("__").size()
                        boost::dynamic_bitset<> bdbs = verges.Encode(
                            generator.Estimate(result->get_double(attr1_name), result->get_double(attr2_name), attr1_name, attr2_name));
                        for (std::size_t indx = 0; indx < bdbs.size(); indx++ ) {
                            encoded_bitset.push_back(bdbs[indx]);
                        }
                    }
                }

                this->emplace_back(std::make_unique<Example>(real_sign, std::move(encoded_bitset)));
            }
        }


        /** default destructor */
        ~Sample() = default;
};

} // namespace vkf
#endif // VKF_SAMPLE_HPP
